from edi_835_parser.elements.identifier import Identifier
from edi_835_parser.elements.payment_method import PaymentMethod
from edi_835_parser.elements.dollars import Dollars
from edi_835_parser.elements.integer import Integer
from edi_835_parser.elements.date import Date
from edi_835_parser.segments.utilities import split_segment


class FinancialInformation:
	identification = 'BPR'

	identifier = Identifier()
	amount_paid = Dollars()
	payment_method = PaymentMethod()
	routing_number = Integer()
	transaction_date = Date()

	def __init__(self, segment: str):
		self.segment = segment
		segment = split_segment(segment)

		self.identifier = segment[0]
		self.amount_paid = segment[2]
		self.payment_method = segment[4]
		self.routing_number = segment[13]
		self.transaction_date = segment[16]

	def __repr__(self):
		return '\n'.join(str(item) for item in self.__dict__.items())


if __name__ == '__main__':
	pass



from typing import Union
from datetime import datetime
from warnings import warn

from edi_835_parser.elements import Element


class Date(Element):

	def parser(self, value: str) -> Union[datetime, str]:
		if len(value) == 10:
			year, month, day, minute, second = [int(value[i:i + 2]) for i in range(0, len(value), 2)]
			return datetime(2000 + year, month, day, minute, second)

		elif len(value) == 8:
			year, month, day = int(value[:4]), int(value[4:6]), int(value[6:])
			return datetime(year, month, day)

		else:
			warn(f'Unable to parse {value} into a datetime')
			return value



from edi_835_parser.elements.identifier import Identifier
from edi_835_parser.elements.entity_code import EntityCode
from edi_835_parser.elements.entity_type import EntityType
from edi_835_parser.elements.identification_code_qualifier import IdentificationCodeQualifier
from edi_835_parser.segments.utilities import split_segment, get_element


class Entity:
	identification = 'NM1'

	identifier = Identifier()
	entity = EntityCode()
	type = EntityType()
	identification_code_qualifier = IdentificationCodeQualifier()

	def __init__(self, segment: str):
		self.segment = segment
		segment = split_segment(segment)

		self.identifier = segment[0]
		self.entity = segment[1]
		self.type = segment[2]
		self.last_name = segment[3]
		self.first_name = get_element(segment, 4)
		self.identification_code_qualifier = get_element(segment, 8)
		self.identification_code = get_element(segment, 9)

	def __repr__(self):
		return '\n'.join(str(item) for item in self.__dict__.items())

	@property
	def name(self) -> str:
		return f'{self.first_name} {self.last_name}'.title()


if __name__ == '__main__':
	pass



from typing import List, Iterable

import pandas as pd

from edi_835_parser.transaction_set.transaction_set import TransactionSet


class TransactionSets:

	def __init__(self, transaction_sets: List[TransactionSet]):
		self.transaction_sets = transaction_sets

	def __iter__(self) -> Iterable[TransactionSet]:
		yield from self.transaction_sets

	def __len__(self) -> int:
		return len(self.transaction_sets)

	def __repr__(self):
		return '\n'.join(str(transaction_set) for transaction_set in self)

	def to_dataframe(self) -> pd.DataFrame:
		data = pd.DataFrame()
		for transaction_set in self:
			data = pd.concat([data, transaction_set.to_dataframe()])

		data = TransactionSets.sort_columns(data)
		return data

	def to_summary_dataframe(self) -> pd.DataFrame:
		data = pd.DataFrame()
		for transaction_set in self:
			data = pd.concat([data, transaction_set.to_summary_dataframe()])

		data = TransactionSets.sort_columns(data)
		return data

	@staticmethod
	def sort_columns(data: pd.DataFrame) -> pd.DataFrame:
		substrings = ['adj', 'ref', 'rem']
		variable_columns = [c for c in data.columns if any(sub_string in c for sub_string in substrings)]
		variable_columns = sorted(variable_columns)

		static_columns = [c for c in data.columns if c not in variable_columns]

		data = data[static_columns + variable_columns]
		return data

	def sum_payments(self) -> float:
		amount = 0
		for transaction_set in self:
			amount += transaction_set.financial_information.amount_paid

		return amount

	def count_claims(self) -> int:
		count = 0
		for transaction_set in self:
			count += len(transaction_set.claims)

		return count

	def count_patients(self) -> int:
		patients = []
		for transaction_set in self:
			for claim in transaction_set.claims:
				patient = claim.patient
				patients.append(patient.identification_code)

		patients = set(patients)
		return len(patients)


from typing import Optional

from edi_835_parser.elements import Element


class AuthorizationInformationQualifier(Element):

	def parser(self, value: str) -> Optional[str]:
		if value == '00':
			value = None

		return value



from edi_835_parser.elements.identifier import Identifier
from edi_835_parser.elements.claim_status import ClaimStatus
from edi_835_parser.elements.dollars import Dollars
from edi_835_parser.segments.utilities import split_segment


class Claim:
	identification = 'CLP'

	identifier = Identifier()
	status = ClaimStatus()
	charge_amount = Dollars()
	paid_amount = Dollars()
	patient_responsibility_amount = Dollars()

	def __init__(self, segment: str):
		self.segment = segment
		segment = split_segment(segment)

		self.identifier = segment[0]
		self.marker = segment[1]
		self.status = segment[2]
		self.charge_amount = segment[3]
		self.paid_amount = segment[4]
		self.patient_responsibility_amount = segment[5]
		# self.claim_filing_indicator_code = segment[6]
		self.payer_claim_control_number = segment[7]

	def __repr__(self):
		return '\n'.join(str(item) for item in self.__dict__.items())


if __name__ == '__main__':
	pass



from edi_835_parser.elements import Element, Code

remark_qualifiers = {
	'HE': 'claim payment'
}


class RemarkQualifier(Element):

	def parser(self, value: str) -> Code:
		description = remark_qualifiers.get(value, None)
		return Code(value, description)



from edi_835_parser.elements.identifier import Identifier
from edi_835_parser.elements.dollars import Dollars
from edi_835_parser.elements.service_code import ServiceCode
from edi_835_parser.elements.service_qualifier import ServiceQualifer
from edi_835_parser.elements.service_modifier import ServiceModifier
from edi_835_parser.elements.integer import Integer
from edi_835_parser.segments.utilities import split_segment, get_element


class Service:
	identification = 'SVC'

	identifier = Identifier()
	charge_amount = Dollars()
	paid_amount = Dollars()
	code = ServiceCode()
	qualifier = ServiceQualifer()
	modifier = ServiceModifier()
	allowed_units = Integer()
	billed_units = Integer()

	def __init__(self, segment: str):
		self.segment = segment
		segment = split_segment(segment)

		self.identifier = segment[0]
		self.code = segment[1]
		self.qualifier = segment[1]
		self.modifier = segment[1]
		self.charge_amount = segment[2]
		self.paid_amount = segment[3]

		# assume unit count of one if unit not provided
		default = 0 if self.paid_amount == 0 else 1
		self.allowed_units = get_element(segment, 5, default=default)

		self.billed_units = get_element(segment, 7, default=self.allowed_units)

	def __repr__(self):
		return '\n'.join(str(item) for item in self.__dict__.items())


if __name__ == '__main__':
	pass



from typing import List, Iterator, Optional
from collections import namedtuple

import pandas as pd

from edi_835_parser.loops.claim import Claim as ClaimLoop
from edi_835_parser.loops.service import Service as ServiceLoop
from edi_835_parser.loops.organization import Organization as OrganizationLoop
from edi_835_parser.segments.utilities import find_identifier
from edi_835_parser.segments.interchange import Interchange as InterchangeSegment
from edi_835_parser.segments.financial_information import FinancialInformation as FinancialInformationSegment
from edi_835_parser.segments.reassociation_trace_number import ReassociationTraceNumber as ReassociationTraceNumberSegment

BuildAttributeResponse = namedtuple('BuildAttributeResponse', 'key value segment segments')


class TransactionSet:

	def __init__(
			self,
			interchange: InterchangeSegment,
			financial_information: FinancialInformationSegment,
			reassociation_trace_number: ReassociationTraceNumberSegment,
			claims: List[ClaimLoop],
			organizations: List[OrganizationLoop],
			file_path: str,
	):
		self.interchange = interchange
		self.financial_information = financial_information
		self.reassociation_trace_number = reassociation_trace_number
		self.claims = claims
		self.organizations = organizations
		self.file_path = file_path

	def __repr__(self):
		return '\n'.join(str(item) for item in self.__dict__.items())

	@property
	def payer(self) -> OrganizationLoop:
		payer = [o for o in self.organizations if o.organization.type == 'payer']
		assert len(payer) == 1
		return payer[0]

	@property
	def payee(self) -> OrganizationLoop:
		payee = [o for o in self.organizations if o.organization.type == 'payee']
		assert len(payee) == 1
		return payee[0]

	def to_dataframe(self) -> pd.DataFrame:
		"""flatten the remittance advice by service to a pandas DataFrame"""
		data = []
		for claim in self.claims:
			for service in claim.services:

				datum = TransactionSet.serialize_service(
					self.financial_information,
					self.reassociation_trace_number,
					self.payer,
					claim,
					service
				)

				for index, adjustment in enumerate(service.adjustments):
					datum[f'adj_{index}_group'] = adjustment.group_code.code
					datum[f'adj_{index}_code'] = adjustment.reason_code.code
					datum[f'adj_{index}_amount'] = adjustment.amount

				for index, reference in enumerate(service.references):
					datum[f'ref_{index}_qual'] = reference.qualifier.code
					datum[f'ref_{index}_value'] = reference.value

				for index, remark in enumerate(service.remarks):
					datum[f'rem_{index}_qual'] = remark.qualifier.code
					datum[f'rem_{index}_code'] = remark.code.code

				data.append(datum)

		return pd.DataFrame(data)

	def to_summary_dataframe(self) -> pd.DataFrame:
		"""flatten the remittance advice by claim to a pandas DataFrame"""
		data = []
		for claim in self.claims:
			datum = {
				'marker': claim.claim.marker,
				'patient': claim.patient.name,
				'payer': self.payer.organization.name,
				'charge_amount': claim.claim.charge_amount,
				'paid_amount': claim.claim.paid_amount,
				'patient_responsibility_amount': claim.claim.patient_responsibility_amount,
				# 'claim_filing_indicator_code': claim.claim_filing_indicator_code,
				'payer_claim_control_number': claim.claim.payer_claim_control_number,
				'payer_classification': str(claim.claim.status.payer_classification),
				'check_number': self.reassociation_trace_number.check_number,
			}

			for service in claim.services:

				for index, adjustment in enumerate(service.adjustments):
					datum[f'adj_{index}_group'] = adjustment.group_code.code
					datum[f'adj_{index}_code'] = adjustment.reason_code.code
					datum[f'adj_{index}_amount'] = adjustment.amount

				for index, reference in enumerate(service.references):
					datum[f'ref_{index}_qual'] = reference.qualifier.code
					datum[f'ref_{index}_value'] = reference.value

				for index, remark in enumerate(service.remarks):
					datum[f'rem_{index}_qual'] = remark.qualifier.code
					datum[f'rem_{index}_code'] = remark.code.code

			data.append(datum)

		return pd.DataFrame(data)

	@staticmethod
	def serialize_service(
			financial_information: FinancialInformationSegment,
			reassociation_trace_number: ReassociationTraceNumberSegment,
			payer: OrganizationLoop,
			claim: ClaimLoop,
			service: ServiceLoop,
	) -> dict:
		# if the service doesn't have a start date assume the service and claim dates match
		start_date = None
		if service.service_period_start:
			start_date = service.service_period_start.date
		elif claim.claim_statement_period_start:
			start_date = claim.claim_statement_period_start.date

		# if the service doesn't have an end date assume the service and claim dates match
		end_date = None
		if service.service_period_end:
			end_date = service.service_period_end.date
		elif claim.claim_statement_period_end:
			end_date = claim.claim_statement_period_end.date

		datum = {
			'marker': claim.claim.marker,
			'patient': claim.patient.name,
			'code': service.service.code,
			'modifier': service.service.modifier,
			'qualifier': service.service.qualifier,
			'allowed_units': service.service.allowed_units,
			'billed_units': service.service.billed_units,
			'transaction_date': financial_information.transaction_date,
			'charge_amount': service.service.charge_amount,
			'allowed_amount': service.allowed_amount,
			'paid_amount': service.service.paid_amount,
			'payer': payer.organization.name,
			'start_date': start_date,
			'end_date': end_date,
			'rendering_provider': claim.rendering_provider.name if claim.rendering_provider else None,
			'payer_classification': str(claim.claim.status.payer_classification),
			'check_number': reassociation_trace_number.check_number,
		}

		return datum

	@classmethod
	def build(cls, file_path: str) -> 'TransactionSet':
		interchange = None
		financial_information = None
		reassociation_trace_number = None
		claims = []
		organizations = []

		with open(file_path) as f:
			file = f.read()

		segments = file.split('~')
		segments = [segment.strip() for segment in segments]

		segments = iter(segments)
		segment = None

		while True:
			response = cls.build_attribute(segment, segments)
			segment = response.segment
			segments = response.segments

			# no more segments to parse
			if response.segments is None:
				break

			if response.key == 'interchange':
				interchange = response.value

			if response.key == 'financial information':
				financial_information = response.value

			if response.key == 'reassociation trace number':
				reassociation_trace_number = response.value

			if response.key == 'organization':
				organizations.append(response.value)

			if response.key == 'claim':
				claims.append(response.value)

		return TransactionSet(interchange, financial_information, reassociation_trace_number, claims, organizations, file_path)

	@classmethod
	def build_attribute(cls, segment: Optional[str], segments: Iterator[str]) -> BuildAttributeResponse:
		if segment is None:
			try:
				segment = segments.__next__()
			except StopIteration:
				return BuildAttributeResponse(None, None, None, None)

		identifier = find_identifier(segment)

		if identifier == InterchangeSegment.identification:
			interchange = InterchangeSegment(segment)
			return BuildAttributeResponse('interchange', interchange, None, segments)

		if identifier == FinancialInformationSegment.identification:
			financial_information = FinancialInformationSegment(segment)
			return BuildAttributeResponse('financial information', financial_information, None, segments)

		if identifier == ReassociationTraceNumberSegment.identification:
			reassociation_trace_number = ReassociationTraceNumberSegment(segment)
			return BuildAttributeResponse('reassociation trace number', reassociation_trace_number, None, segments)

		if identifier == OrganizationLoop.initiating_identifier:
			organization, segments, segment = OrganizationLoop.build(segment, segments)
			return BuildAttributeResponse('organization', organization, segment, segments)

		elif identifier == ClaimLoop.initiating_identifier:
			claim, segments, segment = ClaimLoop.build(segment, segments)
			return BuildAttributeResponse('claim', claim, segment, segments)

		else:
			return BuildAttributeResponse(None, None, None, segments)


if __name__ == '__main__':
	pass


from edi_835_parser.elements.identifier import Identifier
from edi_835_parser.elements.remark_qualifier import RemarkQualifier
from edi_835_parser.elements.remark_code import RemarkCode
from edi_835_parser.segments.utilities import split_segment


class Remark:
	identification = 'LQ'

	identifier = Identifier()
	qualifier = RemarkQualifier()
	code = RemarkCode()

	def __init__(self, segment: str):
		self.segment = segment
		segment = split_segment(segment)

		self.identifier = segment[0]
		self.qualifier = segment[1]
		self.code = segment[2]

	def __repr__(self):
		return '\n'.join(str(item) for item in self.__dict__.items())


if __name__ == '__main__':
	pass



from edi_835_parser.elements.identifier import Identifier
from edi_835_parser.segments.utilities import split_segment


class Address:
	identification = 'N3'

	identifier = Identifier()

	def __init__(self, segment: str):
		self.segment = segment
		segment = split_segment(segment)

		self.identifier = segment[0]
		self.address = segment[1]

	def __repr__(self):
		return '\n'.join(str(item) for item in self.__dict__.items())


if __name__ == '__main__':
	pass



from edi_835_parser.elements.identifier import Identifier
from edi_835_parser.elements.dollars import Dollars
from edi_835_parser.elements.adjustment_group_code import AdjustmentGroupCode
from edi_835_parser.elements.adjustment_reason_code import AdjustmentReasonCode
from edi_835_parser.segments.utilities import split_segment


class ServiceAdjustment:
	identification = 'CAS'

	identifier = Identifier()
	group_code = AdjustmentGroupCode()
	reason_code = AdjustmentReasonCode()
	amount = Dollars()

	def __init__(self, segment: str):
		self.segment = segment
		segment = split_segment(segment)

		self.identifier = segment[0]
		self.group_code = segment[1]
		self.reason_code = segment[2]
		self.amount = segment[3]

	def __repr__(self):
		return '\n'.join(str(item) for item in self.__dict__.items())


if __name__ == '__main__':
	pass



from typing import Optional

from edi_835_parser.elements import Element


class Identifier(Element):

	def __set__(self, obj, value):
		if obj.identification != value:
			raise ValueError('class identifier does not match segment identifier.')

		value = self.parser(value)
		setattr(obj, self.private_name, value)

	def parser(self, value: str) -> Optional[str]:
		return value



from typing import Iterator, Tuple, Optional

from edi_835_parser.segments.organization import Organization as OrganizationSegment
from edi_835_parser.segments.claim import Claim as ClaimSegment
from edi_835_parser.segments.address import Address as AddressSegment
from edi_835_parser.segments.location import Location as LocationSegment
from edi_835_parser.segments.utilities import find_identifier


class Organization:
	initiating_identifier = OrganizationSegment.identification
	terminating_identifiers = [
		ClaimSegment.identification,
		OrganizationSegment.identification,
		'SE'
	]

	def __init__(self, organization: OrganizationSegment = None, location: LocationSegment = None,
				 address: AddressSegment = None):
		self.organization = organization
		self.location = location
		self.address = address

	def __repr__(self):
		return '\n'.join(str(item) for item in self.__dict__.items())

	@classmethod
	def build(cls, current_segment: str, segments: Iterator[str]) -> Tuple[
		'OrganizationSegment', Optional[Iterator[str]], Optional[str]
	]:
		organization = Organization()
		organization.organization = OrganizationSegment(current_segment)
		while True:
			try:
				segment = segments.__next__()
				identifier = find_identifier(segment)

				if identifier == AddressSegment.identification:
					organization.address = AddressSegment(segment)

				elif identifier == LocationSegment.identification:
					organization.location = LocationSegment(segment)

				elif identifier in cls.terminating_identifiers:
					return organization, segments, segment

			except StopIteration:
				return organization, None, None


if __name__ == '__main__':
	pass


from edi_835_parser.elements import Element

organizations = {
	'AV09311993': 'Availity',
	'ZIRMED': 'Zirmed'
}


class Organization(Element):

	def parser(self, value: str) -> str:
		value = value.strip()
		return organizations.get(value, value)



from abc import ABC, abstractmethod


class Element(ABC):

	def __set_name__(self, owner, name):
		self.private_name = '_' + name

	def __get__(self, obj, obj_type=None):
		return getattr(obj, self.private_name)

	def __set__(self, obj, value):
		value = self.parser(value)
		setattr(obj, self.private_name, value)

	@abstractmethod
	def parser(self, value):
		pass


class Code:

	def __init__(self, code: str, description: str):
		self.code = code
		self.description = description

	def __str__(self) -> str:
		return str(self.__dict__)



from typing import Tuple, Iterator, Optional, List
from warnings import warn

from edi_835_parser.segments.service import Service as ServiceSegment
from edi_835_parser.segments.claim import Claim as ClaimSegment
from edi_835_parser.segments.date import Date as DateSegment
from edi_835_parser.segments.reference import Reference as ReferenceSegment
from edi_835_parser.segments.amount import Amount as AmountSegment
from edi_835_parser.segments.service_adjustment import ServiceAdjustment as ServiceAdjustmentSegment
from edi_835_parser.segments.remark import Remark as RemarkSegment
from edi_835_parser.segments.utilities import find_identifier
from edi_835_parser.elements.dollars import Dollars


class Service:
	initiating_identifier = ServiceSegment.identification
	terminating_identifiers = [
		ServiceSegment.identification,
		ClaimSegment.identification,
		'SE'
	]

	def __init__(
			self,
			service: ServiceSegment = None,
			dates: List[DateSegment] = None,
			references: List[ReferenceSegment] = None,
			remarks: List[RemarkSegment] = None,
			amount: AmountSegment = None,
			adjustments: List[ServiceAdjustmentSegment] = None
	):
		self.service = service
		self.dates = dates if dates else []
		self.references = references if references else []
		self.remarks = remarks if remarks else []
		self.amount = amount
		self.adjustments = adjustments if adjustments else []

	def __repr__(self):
		return '\n'.join(str(item) for item in self.__dict__.items())

	@property
	def allowed_amount(self) -> Optional[Dollars]:
		if self.amount:
			if self.amount.qualifier == 'allowed - actual':
				return self.amount.amount

	@property
	def service_date(self) -> Optional[DateSegment]:
		service_date = [d for d in self.dates if d.qualifier == 'service']
		assert len(service_date) <= 1, f'{self.dates}'

		if len(service_date) == 1:
			return service_date[0]

	@property
	def service_period_start(self) -> Optional[DateSegment]:
		service_period_start = [d for d in self.dates if d.qualifier == 'service period start']
		assert len(service_period_start) <= 1, f'{self.dates}'

		if len(service_period_start) == 1:
			return service_period_start[0]
		else:
			return self.service_date

	@property
	def service_period_end(self) -> Optional[DateSegment]:
		service_period_end = [d for d in self.dates if d.qualifier == 'service period end']
		assert len(service_period_end) <= 1

		if len(service_period_end) == 1:
			return service_period_end[0]
		else:
			return self.service_date

	@classmethod
	def build(cls, segment: str, segments: Iterator[str]) -> Tuple['Service', Optional[str], Optional[Iterator[str]]]:
		service = Service()
		service.service = ServiceSegment(segment)

		while True:
			try:
				segment = segments.__next__()
				identifier = find_identifier(segment)

				if identifier == DateSegment.identification:
					date = DateSegment(segment)
					service.dates.append(date)

				elif identifier == AmountSegment.identification:
					service.amount = AmountSegment(segment)

				elif identifier == RemarkSegment.identification:
					remark = RemarkSegment(segment)
					service.remarks.append(remark)

				elif identifier == ReferenceSegment.identification:
					reference = ReferenceSegment(segment)
					service.references.append(reference)

				elif identifier == ServiceAdjustmentSegment.identification:
					service.adjustments.append(ServiceAdjustmentSegment(segment))

				elif identifier in cls.terminating_identifiers:
					return service, segment, segments

				else:
					message = f'Identifier: {identifier} not handled in service loop.'
					warn(message)

			except StopIteration:
				return service, None, None


if __name__ == '__main__':
	pass
